#include "ExportMD3.h"
#include "DiskIO.h"
#include <list>
#include <vector>

BOOL exportSelected;

//////////////////////////////////////////////////////////////////////////////////
// CLASS MySceneEntry
//////////////////////////////////////////////////////////////////////////////////
class MySceneEntry
{
public:
	MySceneEntry(INode *n, Object *o) { node = n; obj = o; next = NULL; }

	INode *			node;
	Object *		obj;
	MySceneEntry *	next;
};

//////////////////////////////////////////////////////////////////////////////////
// CLASS MySceneEnumProc
//////////////////////////////////////////////////////////////////////////////////
class MySceneEnumProc : public ITreeEnumProc
{
public:
	MySceneEnumProc(IScene *scene, TimeValue t, Interface *i);
	~MySceneEnumProc();

	Interface *		i;

	MySceneEntry *	head;
	MySceneEntry *	tail;

	IScene		*	theScene;
	int				count;
	TimeValue		time;
	int				Count() { return count; }
	void			Append(INode *node, Object *obj);
	int				callback( INode *node );
	Box3			Bound();

	MySceneEntry *	operator[](int index);
};

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
MySceneEnumProc::MySceneEnumProc(IScene *scene, TimeValue t, Interface *i)
{
	time		= t;
	theScene	= scene;
	count		= 0;
	head = tail = NULL;
	this->i		= i;

	theScene->EnumTree(this);
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
MySceneEnumProc::~MySceneEnumProc()
{
	while(head)
	{
		MySceneEntry *next = head->next;
		delete head;
		head = next;
	}
	head = tail = NULL;
	count = 0;	
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
int MySceneEnumProc::callback(INode *node)
{
	std::list<std::string>::iterator name_i;
	
	if(exportSelected && node->Selected() == FALSE)
		return TREE_CONTINUE;

	Object *obj = node->EvalWorldState(time).obj;
	
	const char * nodename = node->GetName();

	if(obj->CanConvertToType(triObjectClassID))
	{
		// Is the 'Ignore Bip' option on and this node starts with 'Bip'?
		if( g_ignore_bip && (_tcsncmp( "Bip", nodename, 3 ) == 0) )
			return TREE_CONTINUE;
		
		// Is this a tag?
		if( _tcsncmp( "tag_", nodename, 4 ) == 0 )
		{
			// its a tag - if we want all tags write it out otherwise check if its one we want
			if( !g_only_tags_named )
				Append( node, obj );
			else
			{
				// we want only tags named...
				for( name_i = gl_valid_tags.begin(); name_i != gl_valid_tags.end(); name_i++ )
				{
					if( _tcscmp((*name_i).c_str(), nodename) == 0 )
					{
						Append( node, obj );
						break;
					}
				}
				// ...and the extra special tag_floor if necessary of course.
				if( (g_relative_to_tag == tag_floor) && (_tcscmp("tag_floor", nodename) == 0) )
					Append( node, obj );
			}
		}
		else if( !g_only_objects_named )	// not a tag must be an object - do we want everything?
			Append( node, obj );
		else
		{
			// we want only objects named...
			for( name_i = gl_valid_names.begin(); name_i != gl_valid_names.end(); name_i++ )
			{
				if( _tcsncmp((*name_i).c_str(), nodename, (*name_i).length()) == 0 )
				{
					Append( node, obj );
					break;
				}
			}
		}
	}
	return TREE_CONTINUE;	// Keep on enumeratin'!
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
void MySceneEnumProc::Append(INode *node, Object *obj)
{
	MySceneEntry *entry = new MySceneEntry(node, obj);

	if(tail)
		tail->next = entry;
	
	tail = entry;
	
	if(!head)
		head = entry;
	
	count++;	
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
Box3 MySceneEnumProc::Bound()
{
	Box3 bound;

	bound.Init();
	MySceneEntry *e = head;
	
	ViewExp *vpt = i->GetViewport(NULL);
	
	while(e)
	{
		Box3 bb;
		e->obj->GetWorldBoundBox(time, e->node, vpt, bb);
		bound += bb;
		e = e->next;
	}

	return bound;
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
MySceneEntry *MySceneEnumProc::operator[](int index)
{
	MySceneEntry *e = head;

	while(index)
	{
		e = e->next;
		index--;
	}
	
	return e;
}

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
struct UV_Coord
{
	UV_Coord() : filled( false ) {};

	float	u,v;
	bool	filled;
	int		tvi;
};

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
struct Triangle
{
	union
	{
		struct 
		{
			int a,b,c;
		};
		int e[3];
	};
};

//////////////////////////////////////////////////////////////////////////////////
int	DoExportMD3(const TCHAR *filename,ExpInterface *ei, Interface *gi,
						BOOL suppressPrompts, DWORD options)
{
	int start_time;
	int relative_tag	= -1;	// -1 means relative to no tag at all
	int current_time	= 0;

	char text[256];

	// parse options
#ifdef _MAX2
	exportSelected = 0;
#else
	exportSelected = (options & SCENE_EXPORT_SELECTED) ? TRUE : FALSE;
#endif
	// loop variables
	int i, j;
	std::list<Range>::iterator	range_i;

	//////////////////////////////////////////////////////////////////////////////////
	// Gather info for writing out.

	// Any frames to export specified?
	if( gl_frame_ranges.empty() )
	{
		MessageBox( GetActiveWindow(), "No Frames Specified.", "No Data", MB_OK);
		return TRUE;
	}
	
	start_time = g_ticks_per_frame * (*(gl_frame_ranges.begin())).first;

	// Make sure there are nodes we're interested in!
	// Ask the scene to enumerate all its nodes so we can determine if there are any we can use
	MySceneEnumProc checkScene(ei->theScene, start_time, gi);

	// Any useful nodes?
	if( !checkScene.Count() )
	{
		MessageBox( GetActiveWindow(), "No Data to Export.", "No Data", MB_OK);
		return TRUE;
	}

	// Make a dialog to give the user some feedback on progress.
	HWND progressDialog = CreateDialog( hInstance, MAKEINTRESOURCE(IDD_EXPORTING), GetActiveWindow(), 0 );

	// Build a bounding box for the objects we're exporting
	Box3 bbox = checkScene.Bound();

	std::list<int> lTags;
	std::list<int> lMeshes;

	for( i = 0; i < checkScene.Count(); ++i )
	{
		INode *		node	= checkScene[i]->node;

		// We consider any object whose name starts with 'tag_' to be a tag.
		if( _tcsncmp( "tag_", node->GetName(), 4 ) == 0 )
		{
			// Look for apropriate tags if the user checked the 'relative to tags' box.
			switch( g_relative_to_tag )
			{
			case tag_head:
				if( strcmp( "tag_head", node->GetName() ) == 0 )
					relative_tag = i;
				break;

			case tag_torso:
				if( strcmp( "tag_torso", node->GetName() ) == 0 )
					relative_tag = i;
				break;

			case tag_floor:
				if( strcmp( "tag_floor", node->GetName() ) == 0 )
					relative_tag = i;
				break;

			default:
				break;
			}

			// We wanna write out this tag unless its the extra special tag_floor
			if( strcmp( "tag_floor", node->GetName() ) != 0 )
				lTags.push_back( i );
		}
		else
		{
			bool		needsDel;
			TriObject *	tri		= GetTriObjectFromNode( checkScene[i]->node, start_time, needsDel );
#ifdef _MAX2
			Mesh &		mesh	= tri->Mesh();
#else
			Mesh &		mesh	= tri->GetMesh();
#endif
			Matrix3		tm		= node->GetObjTMAfterWSM( current_time );
			
			// Make sure this isn't an 'empty' mesh.
			if( mesh.numFaces != 0 )
				lMeshes.push_back( i );

			// Delete the working object, if necessary.
			if( needsDel )
				tri->DeleteThis();
		}
	}

	//////////////////////////////////////////////////////////////////////////////////
	// open a file for writing
	//
	FILE *	file;
	long	pos_current;
	char *	subname;

	file = fopen( filename, "wb" );

	if(!file)
		return FALSE;

	// write out sync pattern
	fwrite( "IDP3", 1, 4, file );

	// write out version
	put32( 15, file );


	// Once we have the filename we mangle it.  We search for
	// 'models' and if found we remove the filename to that part.
	// Thus if we have "C:/quake3/baseq3/models/mapobjects/model.md3"
	// It becomes "models/mapobjects/model.md3"
	// otherwise we just leave it.
	subname = strstr( filename, "models" );
	if( !g_smart_paths || subname == NULL )
		fwrite( filename, 68, 1, file );	// full model filename
	else
		fwrite( subname, 68, 1, file );		// filename from 'models' on.
	
	// write out header
	put32( g_total_frames,	file );		// framenum
	put32( lTags.size(),	file );		// tagnum
	put32( lMeshes.size(),	file );		// meshnum
	put32( 0,				file );		// maxskinnum
	put32( 108,				file );		// headersize
	
	long pos_tagstart	= ftell( file );
	put32( 0, file );		// tagstart

	long pos_tagend		= ftell( file );
	put32( 256, file );		// tagend

	long pos_filesize	= ftell( file );
	put32( 512, file );		// filesize

	// write out some per frame object info.
	for( i = 0; i < g_total_frames; i++ )
	{
		putFloat( -1, file );	// min	X
		putFloat( -1, file );	//		Y
		putFloat( -1, file );	//		Z
		putFloat( 1, file );	// max	X
		putFloat( 1, file );	//		Y
		putFloat( 1, file );	//		Z
		putFloat( 0, file );	// pos	X
		putFloat( 0, file );	//		Y
		putFloat( 0, file );	//		Z
		putFloat( 1, file );	// Scale

		fwrite( "(from 3DSMax)", 16, 1, file );	// Creator
	}

	// fill in tagStart
	pos_current = ftell( file );
	fseek( file, pos_tagstart, SEEK_SET );
	put32( pos_current, file );
	fseek( file, pos_current, SEEK_SET );

	// For each frame
		// write out each tag
	Point3	row;
	for( range_i = gl_frame_ranges.begin(); range_i != gl_frame_ranges.end(); range_i++ )
	{
		for( i = (*range_i).first; i <= (*range_i).last; i++ ) // number of frames
		{
			MySceneEnumProc	current_scene(ei->theScene, i * g_ticks_per_frame, gi);

			current_time = current_scene.time;

			sprintf( text, "Writing Tags in Frame %d of %d\n", i, g_total_frames );
			SendDlgItemMessage( progressDialog, IDC_TEXT_EXPORTING, WM_SETTEXT, 0, (LPARAM)text );
	#ifdef _DEBUG
			OutputDebugString( text );
	#endif

			// write out tags
			for( std::list<int>::iterator tag_i = lTags.begin(); tag_i != lTags.end(); tag_i++ )
			{
				INode *		node	= current_scene[*tag_i]->node;
				Matrix3		tm		= node->GetObjTMAfterWSM( current_time );

				fwrite( node->GetName(), 12, 1, file );	// tagname
				fwrite("unknown", 52, 1, file );		// unknown

				// If a relative_tag exists we transform any tags into the
				// relative_tag's space before writing them out.
				if( relative_tag >= 0 )
				{
					Matrix3 scratch	= (current_scene[relative_tag]->node)->GetObjTMAfterWSM( current_time );
					Matrix3 inv		= Inverse( scratch );
					tm *= inv;
				}

				row = tm.GetRow( 3 );

				// tag_floor is special.  We move everything down so its on the floor proper
				if( (g_relative_to_tag == tag_floor) && (relative_tag >= 0) )
					row.z += FLOOR_LEVEL;

				putFloat( row.x, file );	// pos	X
				putFloat( row.y, file );	//		Y
				putFloat( row.z, file );	//		Z

				row = tm.GetRow( 0 );
				putFloat( row.x, file );	// Rotation Matrix
				putFloat( row.y, file );	// row 0
				putFloat( row.z, file );	//

				row = tm.GetRow( 1 );
				putFloat( row.x, file );	//
				putFloat( row.y, file );	// row 1
				putFloat( row.z, file );	//

				row = tm.GetRow( 2 );
				putFloat( row.x, file );	//
				putFloat( row.y, file );	// row 2
				putFloat( row.z, file );	//
			} // tag
		} // frame
	} // ranges

	// fill in tagEnd
	pos_current = ftell( file );
	fseek( file, pos_tagend, SEEK_SET );
	put32( pos_current, file );
	fseek( file, pos_current, SEEK_SET );

	// Constants
	const Point3 x_axis( 1, 0, 0 );
	const Point3 z_axis( 0, 0, 1 );

	// write out meshes
	char mapname[128];

	// reset back to first frame
	{
		MySceneEnumProc scratch(ei->theScene, start_time, gi);
	}

	for( std::list<int>::iterator mesh_i = lMeshes.begin(); mesh_i != lMeshes.end(); mesh_i++ )
	{
		bool		needsDel;

		INode *		node	= checkScene[*mesh_i]->node;

		TriObject *	tri		= GetTriObjectFromNode( node, start_time, needsDel );
#ifdef _MAX2
		Mesh &		mesh	= tri->Mesh();
#else
		Mesh &		mesh	= tri->GetMesh();
#endif
		Texmap *	tex		= 0;
		Mtl *		mtl		= node->GetMtl();
		if( mtl )
		{
			// We have a material.  Get the texture.  If its not a bitmap we don't want it.
			tex = mtl->GetActiveTexmap();
			if( tex && (tex->ClassID() != Class_ID(BMTEX_CLASS_ID, 0x00)) )
				tex = 0;
		}
		
		// start position
		long pos_meshstart = ftell( file );

		// identifier
		fwrite( "IDP3", 1, 4, file );

		// mesh name
		fwrite( node->GetName(), 68, 1, file );

		sprintf( text, "Writing %s Info\n", node->GetName() );
		SendDlgItemMessage( progressDialog, IDC_TEXT_EXPORTING, WM_SETTEXT, 0, (LPARAM)text );
#ifdef _DEBUG
		OutputDebugString( text );
#endif

		put32( g_total_frames,		file );	// framenum

		if( tex )
			put32( 1,				file );	// skinnum
		else
			put32( 0,				file );	// skinnum
	
		long pos_vertexnum = ftell( file );
		put32( 0,					file );	// vertexnum
		
		put32( mesh.getNumFaces(),	file );	// trianglenum

		long pos_trianglestart = ftell( file );
		put32( 0,					file );	// trianglestart

		put32( 108,					file );	// headersize

		long pos_texvecstart = ftell( file );
		put32( 0,					file );	// texvecstart

		long pos_vertexstart = ftell( file );
		put32( 16,					file );	// vertexstart
		
		long pos_meshsize = ftell( file );
		put32( 32,					file );	// meshsize

		if( tex )	// write out a single 'skin'
		{
			// Once we have the filename we mangle it.  We search for
			// 'models' and if found we remove the filename to that part.
			// Thus if we have "C:/quake3/baseq3/models/mapobjects/texture.tga"
			// It becomes "models/mapobjects/texture.tga"
			// otherwise we just leave it.
			strncpy( mapname, ((BitmapTex *)tex)->GetMapName(), 127 );
			subname = strstr( mapname, "models" );
			if( !g_smart_paths || subname == NULL )
				fwrite( mapname, 68, 1, file );	// full texture filename
			else
				fwrite( subname, 68, 1, file );	// filename from 'models' on.
		}

		// Get info for triangles
		std::vector<Triangle>	vTriangles;
		vTriangles.resize( mesh.getNumFaces() );

		for( i = 0; i < mesh.getNumFaces(); i++ )
		{
			vTriangles[i].a = mesh.faces[i].getVert( 0 );
			vTriangles[i].b = mesh.faces[i].getVert( 1 );
			vTriangles[i].c = mesh.faces[i].getVert( 2 );
		}

		// Someplace to store extra vertices.  An extra vertex is simply a reference
		// to the 'real' vertex.  Also keep track of the current index of the extra's.
		std::list<int>	lExtraVerts;
		int				nextVert = mesh.getNumVerts();

		// Write out texture uv for each vertex.
		// In MAX it is possible that there is not a one to one correspondence
		// between texture UVs and vertices.  In that case we try our best
		// and warn the user.
		bool fudged_uv = false;
		std::vector<UV_Coord>	uv_map;
		uv_map.resize( mesh.getNumVerts() );

		// fill in uv_map by going through each tvFace of the mesh and getting
		// the appropriate TVert values.  Unless all the faces aren't textured
		// of course.  In which case we forget it and just make up our own uv
		// mapping with a bunch of (0.5,0.5) coords.
		if( mesh.getNumTVerts() )
		{
			for( i = 0; i < mesh.getNumFaces(); i++ )
			{
				for( j = 0; j < 3; j++ )
				{
					int _v	= mesh.faces[i].getVert( j );
					int _tv	= mesh.tvFace[i].t[j];

					if( !uv_map[_v].filled )
					{
						// this spot not yet taken
						uv_map[_v].u = (mesh.getTVert(_tv)).x;
						uv_map[_v].v = (mesh.getTVert(_tv)).y;
						uv_map[_v].filled = true;
						uv_map[_v].tvi = _tv;
					}
					else
					{
						// that spot already taken!
						// first check to see if the vertex already there is adequate.
						if( _tv == uv_map[_v].tvi )
							continue;

						int _uvm = mesh.getNumVerts();
						bool found_one = false;
						// check if any of the vertices already added are adequate.
						for( std::list<int>::iterator ev_i = lExtraVerts.begin(); ev_i != lExtraVerts.end(); ev_i++ )
						{
							if( *ev_i == _v && 
								uv_map[_uvm].u == mesh.getTVert(_tv).x &&
								uv_map[_uvm].v == mesh.getTVert(_tv).y )
							{
								vTriangles[i].e[j] = _uvm;
								found_one = true;
								break;
							}
							_uvm++;
						}

						// if we didn't find a suitable vertex yet
						// we'll need to add a vertex so that everything is ok.
						if( !found_one )
						{
							UV_Coord new_uv;
							
							fudged_uv = true;
							lExtraVerts.push_back( _v );
							vTriangles[i].e[j] = nextVert;
							new_uv.filled = true;
							new_uv.u = (mesh.getTVert(_tv)).x;
							new_uv.v = (mesh.getTVert(_tv)).y;
							new_uv.tvi = _tv;
							uv_map.push_back( new_uv );
							nextVert++;
						}
					}
				}
			}
		}
		else
		{
			for( i = 0; i < mesh.getNumVerts(); i++ )
			{
				Point3 n = mesh.getVert( i );

				Point3 _elv_t = n; _elv_t.y = 0;
				Point3 _rot_t = n; _rot_t.z = 0;
#ifdef _MAX2
				Point3 _elv_n = Normalize( _elv_t );
				Point3 _rot_n = Normalize( _rot_t );
#else				
				Point3 _elv_n = _elv_t.Normalize();
				Point3 _rot_n = _rot_t.Normalize();
#endif
				float _elv	= DotProd( z_axis, _elv_n );
				float _rot	= DotProd( x_axis, _rot_n );
				
				_elv = acos(_elv) / (2*PI) * 2;

				if( n.y >= 0 )
					_rot = acos(_rot) / (2*PI);
				else
					_rot = (1.0f - (acos(_rot)/(2*PI)));

				uv_map[i].u = _rot;
				uv_map[i].v = _elv;
			}
			if( g_show_warnings )
				MessageBox( GetActiveWindow(), "No UV mapping was found on this object.\n"
												"Writing out the texture coordinates of a\n"
												"spherical mapping for each vertex.", node->GetName(), MB_OK);
		}

		if( fudged_uv && g_show_warnings )
		{
			MessageBox( GetActiveWindow(), "There was not a one to one correspondence\n"
											"between texture vertices and true vertices.\n"
											"UV map may contain errors.", node->GetName(), MB_OK);
		}

		// fill in trianglestart
		pos_current = ftell( file );
		fseek( file, pos_trianglestart, SEEK_SET );
		put32( pos_current - pos_meshstart, file );
		fseek( file, pos_current, SEEK_SET );

		// Write out triangles.
		for( i = 0; i < mesh.getNumFaces(); i++ )
		{
			put32( vTriangles[i].a, file );	// vertex index
			put32( vTriangles[i].c, file );	// for 3 vertices
			put32( vTriangles[i].b, file );	// of triangle
		}

		// fill in texvecstart
		pos_current = ftell( file );
		fseek( file, pos_texvecstart, SEEK_SET );
		put32( pos_current - pos_meshstart, file );
		fseek( file, pos_current, SEEK_SET );

		// Write out UV mapping coords.
		for( i = 0; i < uv_map.size(); i++ )
		{
			putFloat( uv_map[i].u,			file );	// texture coord u,v
			putFloat( 1.0f - uv_map[i].v,	file );	// for vertex
		}

		// fill in vertexstart
		pos_current = ftell( file );
		fseek( file, pos_vertexstart, SEEK_SET );
		put32( pos_current - pos_meshstart, file );
		fseek( file, pos_current, SEEK_SET );

		// fill in vertexnum
		pos_current = ftell( file );
		fseek( file, pos_vertexnum, SEEK_SET );
		put32( nextVert, file );
		fseek( file, pos_current, SEEK_SET );


		// Delete if necessary
		if( needsDel )
			tri->DeleteThis();

		// write out for each frame
			// the position of each vertex
		short x, y, z;
	
		bool no_normals = false;
		
		for( range_i = gl_frame_ranges.begin(); range_i != gl_frame_ranges.end(); range_i++ )
		{
			for( i = (*range_i).first; i <= (*range_i).last; i++ )
			{
				MySceneEnumProc	current_scene(ei->theScene, i * g_ticks_per_frame, gi);

				current_time = current_scene.time;

				INode *		_node	= current_scene[*mesh_i]->node;

				sprintf( text, "Writing Frame %d of %s\n", i, _node->GetName() );
				SendDlgItemMessage( progressDialog, IDC_TEXT_EXPORTING, WM_SETTEXT, 0, (LPARAM)text );
	#ifdef _DEBUG
				OutputDebugString( text );
	#endif

				TriObject *	_tri	= GetTriObjectFromNode( _node, current_time, needsDel );
	#ifdef _MAX2
				Mesh &		_mesh	= _tri->Mesh();
	#else
				Mesh &		_mesh	= _tri->GetMesh();
	#endif
				Matrix3		_tm		= _node->GetObjTMAfterWSM( current_time );

				_mesh.checkNormals( TRUE );

				if( !_mesh.normalsBuilt )
					no_normals = true;

				// If a relative_tag exists we transform vertices into the tags
				// space before writing them out.
				if( relative_tag >= 0 )
				{
					Matrix3 scratch	= (current_scene[relative_tag]->node)->GetObjTMAfterWSM( current_time );
					Matrix3 inv		= Inverse( scratch );
					_tm *= inv;
				}

				// Premake an array with all the appropo verts and norms.
				std::vector<Point3>	allVerts;
				std::vector<Point3> allNorms;

				allVerts.resize( nextVert );
				allNorms.resize( nextVert );

				for( j = 0; j < _mesh.getNumVerts(); j++ )
				{
					allVerts[j] = _mesh.getVert( j );
					if( !no_normals )
						allNorms[j] = _mesh.getNormal( j );
				}

				for( std::list<int>::iterator v_i = lExtraVerts.begin(); v_i != lExtraVerts.end(); v_i++ )
				{
					allVerts[j] = _mesh.getVert( *v_i );
					if( !no_normals )
						allNorms[j] = _mesh.getNormal( *v_i );
					j++;
				}

				for( j = 0; j < nextVert; j++ ) // number of vertices
				{
	#ifdef _MAX2
					Point3 & v = VectorTransform( _tm, allVerts[j] );
	#else
					Point3 & v = _tm.PointTransform( allVerts[j] );
	#endif
					// tag_floor is special.  We move everything down so its on the floor proper
					if( (g_relative_to_tag == tag_floor) && (relative_tag >= 0) )
						v.z += FLOOR_LEVEL;

					x = (short)(v.x * 64.0f);
					y = (short)(v.y * 64.0f);
					z = (short)(v.z * 64.0f);
					
					put16( x, file );
					put16( y, file );
					put16( z, file );

					// Constants
					const Point3 x_axis( 1, 0, 0 );
					const Point3 z_axis( 0, 0, 1 );

					unsigned char _elevation, _rotation;

					if( no_normals )
					{
						put16( 0x0000, file );
					}
					else
					{ 
						Point3 un_n	= allNorms[j];
	#ifdef _MAX2
						Point3 & n	= VectorTransform( _tm, un_n );
	#else
						Point3 & n	= _tm.PointTransform( un_n );
	#endif
						Point3 _elv_t = n; _elv_t.y = 0;
						Point3 _rot_t = n; _rot_t.z = 0;
	#ifdef _MAX2
						Point3 _elv_n = Normalize( _elv_t );
						Point3 _rot_n = Normalize( _rot_t );
	#else
						Point3 _elv_n = _elv_t.Normalize();
						Point3 _rot_n = _rot_t.Normalize();
	#endif
						float _elv	= DotProd( z_axis, _elv_n );
						float _rot	= DotProd( x_axis, _rot_n );
						
						_elevation	= (unsigned char)( acos(_elv) / (2*PI) * 255 );

						if( n.y >= 0 )
							_rotation	= (unsigned char)( acos(_rot) / (2*PI) * 255 );
						else
							_rotation	= (unsigned char)( (1.0f - (acos(_rot)/(2*PI))) * 255 );
						
						putc( _elevation,	file );	// normal elevation angle: 0 up, 63 sideways and 80 down.
						putc( _rotation,	file );	// normal rotation: 0 straight along X, 40 straight along Y, 7f opposite X, c0 opposite Y
					}
				} // vertices

				// Delete the working object, if necessary.
				if( needsDel )
					_tri->DeleteThis();

			} // frames
		} // ranges

		if( no_normals && g_show_warnings )
		{
			MessageBox( GetActiveWindow(), "Object does not have normals contructed.\n"
											"Enviroment maps will be wrong.", node->GetName(), MB_OK);
		}

		// fill in meshsize
		pos_current = ftell( file );
		fseek( file, pos_meshsize, SEEK_SET );
		put32( pos_current - pos_meshstart, file );
		fseek( file, pos_current, SEEK_SET );

		// reset back to first frame
		{
			MySceneEnumProc scratch(ei->theScene, start_time, gi);
		}

	} // mesh_i


	// fill in filesize
	pos_current = ftell( file );
	fseek( file, pos_filesize, SEEK_SET );
	put32( pos_current, file );
	fseek( file, pos_current, SEEK_SET );

	fclose(file);

	//////////////////////////////////////////////////////////////////////////////////
	// Cleanup.
	if( progressDialog )
		DestroyWindow( progressDialog );
	
	gl_valid_names.clear();
	gl_valid_tags.clear();
	gl_frame_ranges.clear();

	return TRUE;
}
	

